/***************************************************************************
Copyright (c) 2006-2011, Armin Biere, Johannes Kepler University.
Copyright (c) 2022, Daniela Kaufmann, Johannes Kepler University.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#define USAGE \
"usage: aigdd [options] <src> <dst> <run> [ <run-options> ...]\n" \
"\n" \
"This 'delta debugger' for AIGs has the following options:\n" \
"\n" \
"[options]:\n" \
"  -h          prints this command line option summary\n" \
"  -v<i>       different levels of verbosity(min i=1, max i=3, default -v1) \n" \
"  -o          allows that also outputs of the AIG may be removed" \
"  -r          reencode and remove holes even if <dst> is in ASCII format\n" \
"  -so i j     slicing of multiplier circuits by setting the 'i' MSBs of each input bitvector\n" \
"              and 'j' most significant outputs to zero, ordering of inputs is ascending\n" \
"  -si i j     slicing of multiplier circuits by setting the 'i' MSBs of each input bitvector\n" \
"              and 'j' most significant outputs to zero, ordering of inputs is interleaved\n" \
"\n" \
"\n" \
"  <src>       source file in AIGER format\n" \
"  <dst>       destination file in AIGER format\n" \
"  <run>       executable\n" \
"\n" \
"The idea is that '<run> <src>' produces a fault, e.g. you know that\n" \
"there is a big AIG saved in '<src>' which produces a wrong behaviour\n" \
"when given as argument to the program '<run>'.\n" \
"\n" \
"You can now use 'aigdd' to produce a copy '<dst>' of '<src>' in which\n" \
"as many literals as possible are removed while still producing\n" \
"the same exit code when running '<run> <dst>'.  Literals are actually\n" \
"removed by assigning them to a constant.  This in effect removes inputs,\n" \
"latches and ANDs.  The number of outputs is currently not changed, but\n" \
"individual outputs are set to constants.\n" \
"\n" \
"If '<dst>' is an AIG in ASCII format, by specifying a '.aag' extension,\n" \
"then the 'holes' left by removed literals are not squeezed out, while\n" \
"in the binary this is enforced.\n" \
"\n" \
"As a typical example consider that you have a new structural SAT solver\n" \
"'solve' that reads AIGs.  On one AIG it fails with an assertion\n" \
"failure.  You save this AIG in '/tmp/fail.aig'.  To shrink\n" \
"this file while still producing a failure you could just use\n" \
"\n" \
"  aigdd /tmp/fail.aig /tmp/shrunken.aig solve\n" \
"\n" \
"Unless your solver produces the same exit code for a correct run, this\n" \
"should give a compact easier to analyze AIG in '/tmp/shrunken.aig'.\n" \
"\n" \
"The '<dst>' will always have the smallest failure inducing input sofar.\n" \

#include "../includes/aiger.h"

#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/wait.h>           /* for WEXITSTATUS */
/*------------------------------------------------------------------------*/
static aiger *src;
static const char *dst_name;
static char tmp_name[80];
static unsigned *unstable;
static unsigned *stable;
static char * fixed;
static char * outputs;
static char * bad;
static char * constraints;
static char * justice;
static char * fairness;
static int verbose = 1;
static int reencode = 0;
static int runs;


/*------------------------------------------------------------------------*/
static void msg(int level, const char *fmt, ...) {
  va_list ap;
  if (verbose < level) return;
  fputs ("[aigdd2] ", stdout);
  va_start (ap, fmt);
  vfprintf (stdout, fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stderr);
}

/*------------------------------------------------------------------------*/
static void die (const char *fmt, ...) {
  va_list ap;
  fputs ("*** [aigdd2] ERROR ", stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

/*------------------------------------------------------------------------*/
static unsigned deref(unsigned lit) {
  unsigned idx = lit / 2;
  unsigned sign = lit & 1;
  unsigned tmp, res, tmp0, tmp1;
  aiger_and * and;

  tmp = unstable[idx];

  if (tmp == 2 * idx) {
    if (!fixed [idx]) {
	    and = aiger_is_and (src, 2 * idx);

	    if (and){
	      tmp0 = deref(and->rhs0);
	      tmp1 = deref(and->rhs1);

	      if (!tmp0 || !tmp1) tmp = 0;
	      else if (tmp0 == 1) tmp = tmp1;
	      else if (tmp1 == 1)	tmp = tmp0;
	    }
	  }
  } else {
    tmp = deref(tmp);
  }

  unstable[idx] = tmp;
  fixed[idx] = 1;
  res = tmp ^ sign;

  return res;
}

/*------------------------------------------------------------------------*/
static void write_unstable_input(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_inputs; i++) {
    aiger_symbol* symbol = src->inputs + i;
    unsigned lit = symbol->lit;
    if (deref(lit) != lit) continue;
    aiger_add_input (dst, lit, symbol->name);
    if (symbol->reset) aiger_add_reset (dst, lit, symbol->reset);
  }
}

/*------------------------------------------------------------------------*/
static void write_unstable_latch(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_latches; i++) {
    aiger_symbol* symbol = src->latches + i;
    unsigned lit = symbol->lit;
    if (deref(lit) != lit) continue;
    aiger_add_latch (dst, lit, deref(symbol->next), symbol->name);

    if (symbol->reset <= 1) {
      aiger_add_reset (dst, lit, symbol->reset);
    } else {
      assert (symbol->reset == lit);
      aiger_add_reset (dst, lit, lit);
    }
  }
}

/*------------------------------------------------------------------------*/
static void write_unstable_and(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_ands; i++) {
    aiger_and* and = src->ands + i;

    if (deref(and->lhs) == and->lhs) {
      aiger_add_and(dst, and->lhs, deref(and->rhs0), deref(and->rhs1));
    }
  }
}

/*------------------------------------------------------------------------*/
static void write_unstable_output(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_outputs; i++) {
    if (!outputs[i]) continue;
    aiger_symbol* symbol = src->outputs + i;
    aiger_add_output(dst, deref(symbol->lit), symbol->name);
  }
}

/*------------------------------------------------------------------------*/
static void write_unstable_bad(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_bad; i++) {
    if (!bad[i]) continue;
    aiger_symbol* symbol = src->bad + i;
    aiger_add_bad(dst, deref(symbol->lit), symbol->name);
  }
}

/*------------------------------------------------------------------------*/
static void write_unstable_constraint(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_constraints; i++) {
    if (!constraints[i]) continue;
    aiger_symbol* symbol = src->constraints + i;
    aiger_add_constraint(dst, deref(symbol->lit), symbol->name);
  }
}

/*------------------------------------------------------------------------*/
static void write_unstable_justice(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_justice; i++) {
    if (!justice[i]) continue;
    aiger_symbol* symbol = src->justice + i;
    unsigned * lits = malloc(symbol->size * sizeof *lits);

    for (unsigned j = 0; j < symbol->size; j++) {
      lits[j] = deref(symbol->lit);
    }

    aiger_add_justice(dst, symbol->size, lits, symbol->name);
    free (lits);
  }
}

/*------------------------------------------------------------------------*/
static void write_unstable_fairness(const char * name, aiger *dst){
  for (unsigned i = 0; i < src->num_fairness; i++) {
    if (!fairness[i]) continue;
    aiger_symbol* symbol = src->fairness + i;
    aiger_add_fairness(dst, deref(symbol->lit), symbol->name);
  }

}

/*------------------------------------------------------------------------*/
static void write_unstable (const char * name) {
  int print_progress = (name == dst_name);
  memset (fixed, 0, src->maxvar + 1);

  aiger *dst = aiger_init ();
  write_unstable_input(name, dst);
  write_unstable_latch(name, dst);
  write_unstable_and(name, dst);
  write_unstable_output(name, dst);

  write_unstable_bad(name, dst);
  write_unstable_constraint(name, dst);
  write_unstable_justice(name, dst);
  write_unstable_fairness(name, dst);

  assert (!aiger_check(dst));

  if (reencode) aiger_reencode(dst);

  unlink (name);
  if (!aiger_open_and_write_to_file(dst, name)) {
    die ("failed to write '%s'", name);
  }

  if (print_progress) {
    assert (name == dst_name);
    msg(2, "wrote '%s' MILOABCJF %u %u %u %u %u %u %u %u %u",
        name,
	      dst->maxvar,
	      dst->num_inputs,
	      dst->num_latches,
	      dst->num_outputs,
	      dst->num_ands,
	      dst->num_bad,
	      dst->num_constraints,
	      dst->num_justice,
	      dst->num_fairness);
  } else {
    assert (name == tmp_name);
  }

  aiger_reset (dst);
}

/*------------------------------------------------------------------------*/
static void copy_stable_to_unstable_and_write_aig_to_dst_name (void) {
  memcpy(unstable, stable, sizeof (unstable[0]) * (src->maxvar + 1));
  msg(2, "writing '%s'", dst_name);
  write_unstable (dst_name);
}

/*------------------------------------------------------------------------*/
#define CMDSUFFIX " 1>/dev/null 2>/dev/null"
/*------------------------------------------------------------------------*/
static char * strapp (char * str, const char * suffix) {
  return strcat (realloc (str, strlen (str) + strlen (suffix) + 1), suffix);
}

/*------------------------------------------------------------------------*/
static int run (const char * cmd, const char * name) {
  int res;
  char * fullcmd = strdup (cmd);
  fullcmd = strapp (fullcmd, " ");
  fullcmd = strapp (fullcmd, name);
  fullcmd = strapp (fullcmd, CMDSUFFIX);
  runs++;
  msg(3, "full command '%s'", fullcmd);
  res = system (fullcmd);
  free (fullcmd);
  return WEXITSTATUS (res);
}

/*------------------------------------------------------------------------*/
static int write_and_run_unstable (const char * cmd) {
  write_unstable (tmp_name);
  return run (cmd, tmp_name);
}

/*------------------------------------------------------------------------*/
static int min (int a, int b) {
  return a < b ? a : b;
}

static void copy_stable_to_unstable(unsigned min, unsigned max){
  for (unsigned j = min; j <= max; j++) {
    unstable[j] = stable[j];
  }
}

static void copy_unstable_to_stable(unsigned min, unsigned max){
  for (unsigned j = min; j <= max; j++) {
    stable[j] = unstable[j];
  }
}

static unsigned count_stable(unsigned min, unsigned max){
  unsigned count = 0;
  for (unsigned j = min; j <= max; j++) {
    if (stable[j]) count++;
  }
  return count;
}

/*------------------------------------------------------------------------*/
static void allocate_and_init_global_var (const char* src_name){
  const char* err;

  src = aiger_init ();
  if ((err = aiger_open_and_read_from_file (src, src_name)))
    die ("%s: %s", src_name, err);

  stable   = malloc (sizeof (stable[0]) * (src->maxvar + 1));
  unstable = malloc (sizeof (unstable[0]) * (src->maxvar + 1));
  fixed       = malloc (src->maxvar + 1);
  outputs     = malloc (src->num_outputs + sizeof *outputs);
  bad         = malloc (src->num_bad + sizeof *bad);
  constraints = malloc (src->num_constraints + sizeof *constraints);
  justice     = malloc (src->num_justice + sizeof *justice);
  fairness    = malloc (src->num_fairness + sizeof *fairness);

  for (unsigned i = 0; i < src->maxvar+1;        i++) stable[i] = 2 * i;
  for (unsigned i = 0; i < src->num_outputs;     i++) outputs[i] = 1;
  for (unsigned i = 0; i < src->num_bad;         i++) bad[i] = 1;
  for (unsigned i = 0; i < src->num_constraints; i++) constraints[i] = 1;
  for (unsigned i = 0; i < src->num_justice;     i++) justice[i] = 1;
  for (unsigned i = 0; i < src->num_fairness;    i++) fairness[i] = 1;
}

/*------------------------------------------------------------------------*/
static int init_dd (char* cmd, const char* src_name){
  // Execute "cmd" with given original data
  int expected = run (cmd, src_name);
  msg(0, "'%s %s' returns %d", cmd, src_name, expected);

  // Execute "cmd" with copied original data
  copy_stable_to_unstable_and_write_aig_to_dst_name ();
  int res = run (cmd, dst_name);
  msg(1, "'%s %s' returns %d", cmd, dst_name, res);

  // Error if sthg went wrong while copying data
  if (res != expected) {
    die ("exec code on polished copy '%s' of '%s' differs (%d instead of %d)",
         dst_name, src_name, res, expected);
  }

  sprintf (tmp_name, "/tmp/aigdd%d.aig", getpid ());
  msg(1, "using temporary file '%s'", tmp_name);

  return expected;
}

/*------------------------------------------------------------------------*/
static void dd_aig(char* cmd, int expected){
  for (int delta = src->maxvar; delta; delta = (delta == 1) ? 0 : (delta + 1) / 2) {
    unsigned i = 1;

    do
    {
	    unsigned last = min (i + delta - 1, src->maxvar);
	    int outof = last - i + 1;

      copy_stable_to_unstable(1, i-1);
	    for (unsigned j = i; j <= last; j++) {
        unstable[j] = 0; /* always favor 'zero' */
      }

      unsigned changed = count_stable(i, last);
	    if (changed) {
        copy_stable_to_unstable(i + delta, src->maxvar);

	      if (write_and_run_unstable (cmd) == expected) {
		      msg(1, "[%d,%d] set to 0 (%d out of %d)", i, last, changed, outof);

          copy_unstable_to_stable(i, last);
		      copy_stable_to_unstable_and_write_aig_to_dst_name ();

        } else {		/* try setting to 'one' */
		      msg(3, "[%d,%d] can not be set to 0 (%d out of %d)", i, last, changed, outof);

          copy_stable_to_unstable(1, i-1);
		      for (unsigned j = i; j <= last; j++) {
            unstable[j] = stable[j] ? 1 : 0;
          }
		      copy_stable_to_unstable(i + delta, src->maxvar);

		      if (write_and_run_unstable (cmd) == expected) {
			      msg(1, "[%d,%d] set to 1 (%d out of %d)", i, last, changed, outof);

            copy_unstable_to_stable(i, last);
			      copy_stable_to_unstable_and_write_aig_to_dst_name ();
			    } else {
			      msg(3, "[%d,%d] can neither be set to 1 (%d out of %d)",
			        i, last, changed, outof);
          }

        }
	    } else {
	      msg(3, "[%d,%d] stabilized to 0", i, last);
      }

	    i += delta;
	  } while (i <= src->maxvar);
  }
  copy_stable_to_unstable_and_write_aig_to_dst_name ();
}

/*------------------------------------------------------------------------*/
static void dd_special (char* name, char* arr, unsigned len, char* cmd, int expected){
  int changed = 0;

  for (unsigned i = 0; i < len; i++) {
    assert(arr[i]);
    arr[i] = 0;
    if (write_and_run_unstable(cmd) == expected){
      msg(2, "removed %s %d", name, i);
      changed++;
    } else {
      msg(3, "can not remove %s %d", name, i);
      arr[i] = 1;
    }
  }

  msg(1, "removed %d %ss", changed, name);
  copy_stable_to_unstable_and_write_aig_to_dst_name ();
}

/*------------------------------------------------------------------------*/
static void dd_slicing(char* cmd, int expected, int slicing, int ibw, int obw){
  if(slicing == 1) msg(1, "slicing, inputs interleaved");
  if(slicing == 2) msg(1, "slicing, inputs ordered");

  msg(2, "each iteration slices %i inputs and %i outputs", ibw, obw);

  int inp_bitw = src->num_inputs;
  msg(1, "%i %i ", inp_bitw, ibw);

  if(slicing == 2 && inp_bitw % ibw != 0){
    die("ordered slicing not possible, number of inputs not divisible by %i", ibw );
  }
  int intervall = inp_bitw / ibw;
  int out_bitw = src->num_outputs;
  int max_bitw = src->num_inputs >> 1;
  int bitw = max_bitw;

  while(inp_bitw > ibw && out_bitw > obw){
    copy_stable_to_unstable(1, src->maxvar);

    if(slicing == 1){
      for(int i = 1; i<= ibw; i++){
        unstable[(inp_bitw -i)/2 ] = 0;
      }
    } else {
      for(int i = 0; i< ibw; i++){
        unstable[(inp_bitw + 2*i*intervall)/2 ] = 0;
      }
    }

    for(int j = 1; j<= obw; j++){
      outputs[out_bitw-j] = 0;
    }

    if (write_and_run_unstable (cmd) == expected) {
      if(slicing == 1){
        for(int i = 1; i<= ibw; i++){
          msg(3, "input %i set to 0",(inp_bitw -i)/2 );
        }
      } else {
        for(int i = 0; i< ibw; i++){
          msg(3, "input %i set to 0", (inp_bitw + 2*i*intervall)/2);
        }
      }


      for(int j = 1; j<= obw; j++){
        msg(3, "output %i removed", out_bitw-j);
      }
      copy_unstable_to_stable(1, src->maxvar);
      copy_stable_to_unstable_and_write_aig_to_dst_name ();
    } else {
      copy_stable_to_unstable(1, src->maxvar);
      if(slicing == 1){
        for(int i = 1; i<= ibw; i++){
          unstable[(inp_bitw -i)/2 ] = 1;
        }
      } else {
        for(int i = 0; i< ibw; i++){
          unstable[(inp_bitw + 2*i*intervall)/2 ] = 1;
        }
      }


      for(int j = 1; j<= obw; j++){
        outputs[out_bitw-j] = 1;
      }
      msg(1, "slicing stabilized");
      msg(1, "removed %i inputs", src->num_inputs - inp_bitw);
      msg(1, "removed %i outputs",src->num_outputs - out_bitw);
      break;
    }

    out_bitw = out_bitw - obw;
    inp_bitw = inp_bitw - ibw;
  }





  copy_stable_to_unstable_and_write_aig_to_dst_name ();
}

/*------------------------------------------------------------------------*/
static void count_removed_literals(){
    int changed = 0;
    for (unsigned i = 1; i <= src->maxvar; i++) {
      if (stable[i] <= 1) changed++;
    }

    msg(0, "%.1f%% literals removed (%d out of %d)",
         src->maxvar ? changed * 100.0 / src->maxvar : 0, changed, src->maxvar);
}

/*------------------------------------------------------------------------*/
static void free_and_reset_all_global_var (){
  free (fairness);
  free (justice);
  free (constraints);
  free (bad);
  free (stable);
  free (unstable);
  free (fixed);
  aiger_reset (src);
  unlink (tmp_name);
}

/*------------------------------------------------------------------------*/
int main (int argc, char **argv) {
  char * cmd;
  const char * src_name = dst_name = cmd = 0;

  int debug_outputs = 0;
  int slicing = 0;
  int ibw = 0;
  int obw = 0;

  for (int i = 1; i < argc; i++) {

    if (!src_name && !strcmp (argv[i], "-h")) {
	    fprintf (stderr, USAGE);
	    exit (0);
    } else if (!src_name && !strcmp(argv[i], "-v0")) { verbose = 0;
    } else if (!src_name && !strcmp(argv[i], "-v1")) { verbose = 1;
    } else if (!src_name && !strcmp(argv[i], "-v2")) { verbose = 2;
    } else if (!src_name && !strcmp(argv[i], "-v3")) { verbose = 3;
    } else if (!src_name && !strcmp(argv[i], "-o"))  { debug_outputs = 1;
    } else if (!src_name && !strcmp(argv[i], "-so"))  {
      if(i > argc-3) die("too few values for option '-so' given (see '-h')");
      slicing = 2;
      ibw = atoi(argv[i+1]);
      obw = atoi(argv[i+2]);
      i = i+2;
    } else if (!src_name && !strcmp(argv[i], "-si"))  {
      if(i > argc-3) die("too few values for option '-so' given (see '-h')");
      slicing = 1;
      ibw = atoi(argv[i+1]);
      obw = atoi(argv[i+2]);
      i = i+2;
    } else if (!src_name && !strcmp (argv[i], "-r")) { reencode = 1;
    } else if (src_name && dst_name && cmd) {
	    cmd = strapp (strapp (cmd, " "), argv[i]);
    } else if (dst_name) {
	    cmd = strdup (argv[i]);
    } else if (src_name) {
      dst_name = argv[i];
    } else {
	    src_name = argv[i];
    }
  }

  if (!src_name || !dst_name) die ("expected exactly two files");
  if (!cmd) die ("name of executable missing");

  allocate_and_init_global_var(src_name);

  int expected = init_dd(cmd, src_name);

  if(slicing) dd_slicing(cmd, expected, slicing, ibw, obw);

  dd_aig(cmd, expected);
  dd_special("bad state",              bad,         src->num_bad,         cmd, expected);
  dd_special("environment constraint", constraints, src->num_constraints, cmd, expected);
  dd_special("justice constraint",     justice,     src->num_justice,     cmd, expected);
  dd_special("fairness constraint",    fairness,    src->num_fairness,    cmd, expected);
  if (debug_outputs) dd_special("output", outputs,  src->num_outputs,     cmd, expected);



  count_removed_literals();

  free_and_reset_all_global_var();
  free (cmd);

  return 0;
}
